// @ts-check
import type { ESLintConfig } from 'eslint-define-config';

export function extendConfig({
  plugins,
  settings,
  rules,
  ...config
}: ESLintConfig): ESLintConfig {
  return {
    root: true,
    parser: '@typescript-eslint/parser',
    plugins: ['@typescript-eslint', 'import', ...(plugins ?? [])],
    env: { node: true, browser: true },
    reportUnusedDisableDirectives: true,
    parserOptions: { project: ['./tsconfig.json'] },
    extends: [
      'eslint:recommended',
      'prettier',
      'plugin:@typescript-eslint/recommended',
      'plugin:import/errors',
      'plugin:import/typescript',
      'plugin:react/recommended',
      'plugin:react-hooks/recommended',
      'plugin:jsx-a11y/recommended',
    ],
    settings: {
      'import/parsers': {
        '@typescript-eslint/parser': ['.ts', '.tsx'],
      },
      'import/core-modules': ['node:test'],
      'import/resolver': {
        typescript: {
          alwaysTryTypes: true,
          project: './tsconfig.json',
        },
      },
      react: {
        version: 'detect',
      },
      ...settings,
    },
    rules: {
      'no-duplicate-imports': 'error',
      'no-restricted-imports': [
        'error',
        {
          paths: [
            {
              name: 'crypto',
              importNames: ['webcrypto'],
              message: 'Use global `crypto` instead',
            },
          ],
        },
      ],
      'no-restricted-globals': ['error', 'event', 'name'],
      '@typescript-eslint/ban-ts-comment': 'off',
      '@typescript-eslint/consistent-type-imports': [
        'error',
        { disallowTypeAnnotations: false, fixStyle: 'inline-type-imports' },
      ],
      '@typescript-eslint/ban-types': [
        'error',
        {
          extendDefaults: false,
          types: {
            String: { message: 'Use string instead', fixWith: 'string' },
            Number: { message: 'Use number instead', fixWith: 'number' },
            Boolean: { message: 'Use boolean instead', fixWith: 'boolean' },
            Symbol: { message: 'Use symbol instead', fixWith: 'symbol' },
          },
        },
      ],
      '@typescript-eslint/explicit-function-return-type': 'off',
      '@typescript-eslint/explicit-module-boundary-types': 'off',
      '@typescript-eslint/no-empty-function': 'off',
      '@typescript-eslint/no-explicit-any': 'off',
      '@typescript-eslint/no-namespace': 'off',
      '@typescript-eslint/no-non-null-assertion': 'off',
      '@typescript-eslint/no-use-before-define': 'off',
      '@typescript-eslint/no-var-requires': 'off',
      '@typescript-eslint/triple-slash-reference': 'off',
      '@typescript-eslint/no-empty-interface': 'off',
      '@typescript-eslint/no-unnecessary-type-assertion': 'error',
      '@typescript-eslint/no-unused-vars': [
        'error',
        { ignoreRestSiblings: true, varsIgnorePattern: '^_' },
      ],
      'arrow-body-style': ['error', 'as-needed'],
      'class-methods-use-this': 'off',
      complexity: ['warn', { max: 100 }],
      curly: ['error', 'multi-line', 'consistent'],
      eqeqeq: ['error', 'smart'],
      'no-async-promise-executor': 'off',
      'no-case-declarations': 'off',
      'no-console': 'warn',
      'no-constant-condition': ['error', { checkLoops: false }],
      'no-debugger': 'off',
      'no-empty': ['error', { allowEmptyCatch: true }],
      'no-inner-declarations': 'off',
      'no-lonely-if': 'error',
      'no-template-curly-in-string': 'error',
      'no-var': 'error',
      'import/export': 'off',
      'import/order': ['error', { groups: ['builtin', 'external'] }],
      'object-shorthand': ['error', 'always', { ignoreConstructors: true }],
      'one-var': ['error', { var: 'never', let: 'never' }],
      'prefer-arrow-callback': 'error',
      'prefer-const': ['error', { destructuring: 'all' }],
      'prefer-destructuring': [
        'warn',
        { AssignmentExpression: { array: false, object: false } },
      ],
      'prefer-object-spread': 'error',
      'prefer-rest-params': 'warn',
      'prefer-spread': 'warn',
      'quote-props': ['error', 'as-needed'],
      'spaced-comment': ['error', 'always', { markers: ['/'] }],
      'sort-imports': ['warn', { ignoreDeclarationSort: true }],
      yoda: ['error', 'never', { exceptRange: true }],
      'react/display-name': 'off',
      'react/no-children-prop': 'error',
      'react/prop-types': 'off',
      'react/react-in-jsx-scope': 'off',
      'react/no-unknown-property': ['error', { ignore: ['css'] }],
      ...rules,
    },
    ...config,
  };
}
